#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

// This kernel fuses the linear transformation (matrix multiplication + bias addition) with a max pooling and summation stage.
// Each block processes one sample from the batch. The computed output vector is stored in shared memory to avoid writing
// an intermediate result to global memory. We use warp-level primitives (__shfl_down_sync) to efficiently reduce 
// the partial sums from the max pooling operation.

__global__ void fused_divergence_kernel(
    const float* __restrict__ x,
    const float* __restrict__ weight,
    const float* __restrict__ bias,
    float* __restrict__ output,
    int B,          // batch size
    int inF,        // number of input features
    int outF,       // number of output features
    int kernel_size,
    float scale_factor
) {
    const int warp_size = 32;
    const int lane_id = threadIdx.x % warp_size;
    const int warp_id = threadIdx.x / warp_size;
    const int block_threads = blockDim.x;
    const int warps_per_block = block_threads / warp_size;

    // Each block processes one sample in the batch
    int b = blockIdx.x;
    if (b >= B) return;

    // Use shared memory to store the linear transformation results and warp partial sums.
    // Assumed shared memory size: (outF + warps_per_block) floats
    extern __shared__ float sdata[];
    float* linear_out = sdata;             // holds linear transformation results (size: outF)
    float* warp_sum   = &sdata[outF];        // holds partial pooling sums from each warp

    // Partition the output features among warps
    int features_per_warp = (outF + warps_per_block - 1) / warps_per_block;
    int start_feature = warp_id * features_per_warp;
    int end_feature = min(start_feature + features_per_warp, outF);

    // Compute the linear transformation: out[j] = bias[j] + dot(x[b], weight[j])
    // Each warp processes its assigned chunk with vectorized thread-level parallelism
    for (int j = start_feature + lane_id; j < end_feature; j += warp_size) {
        float sum = bias[j];
        const float* w_row = weight + j * inF;  // row for output feature j
        const float* x_row = x + b * inF;
        #pragma unroll 4
        for (int i = 0; i < inF; i++) {
            sum += x_row[i] * w_row[i];
        }
        linear_out[j] = sum;
    }
    __syncthreads();

    // Compute pooling segments
    // The pooling is performed over segments of size 'kernel_size' along the output feature dimension.
    int pooled_len = (outF >= kernel_size) ? (1 + (outF - kernel_size) / kernel_size) : 0;

    // Each thread processes multiple segments in a round-robin fashion
    float thread_partial_sum = 0.0f;
    for (int seg = threadIdx.x; seg < pooled_len; seg += block_threads) {
        int seg_start = seg * kernel_size;
        float seg_max = linear_out[seg_start];
        // Compute the max in this pooling segment
        for (int k = 1; k < kernel_size; k++) {
            int idx = seg_start + k;
            if (idx < outF) {
                float val = linear_out[idx];
                seg_max = (val > seg_max) ? val : seg_max;
            }
        }
        thread_partial_sum += seg_max;
    }

    // Warp-level reduction of the partial sums (using shuffle operations)
    for (int offset = warp_size/2; offset > 0; offset /= 2) {
        thread_partial_sum += __shfl_down_sync(0xffffffff, thread_partial_sum, offset);
    }

    // First lane of each warp writes its result to shared memory
    if (lane_id == 0) {
        warp_sum[warp_id] = thread_partial_sum;
    }
    __syncthreads();

    // Final reduction among warp results by the first warp
    if (warp_id == 0) {
        float block_sum = (lane_id < warps_per_block) ? warp_sum[lane_id] : 0.0f;
        for (int offset = warp_size/2; offset > 0; offset /= 2) {
            block_sum += __shfl_down_sync(0xffffffff, block_sum, offset);
        }
        if (lane_id == 0) {
            output[b] = block_sum * scale_factor;
        }
    }
}


at::Tensor forward(
    at::Tensor x,
    int64_t kernel_size,
    double scale_factor,
    at::Tensor weight,
    at::Tensor bias
) {
    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(weight.is_cuda(), "weight must be a CUDA tensor");
    TORCH_CHECK(bias.is_cuda(), "bias must be a CUDA tensor");
    TORCH_CHECK(x.dim() == 2, "x must have shape (batch_size, in_features)");
    TORCH_CHECK(weight.dim() == 2, "weight must have shape (out_features, in_features)");
    TORCH_CHECK(bias.dim() == 1, "bias must have shape (out_features)");

    const auto B = x.size(0);
    const auto inF = x.size(1);
    const auto outF = weight.size(0);

    auto out = torch::empty({B}, x.options());

    // Configure kernel launch parameters
    const int threads_per_block = 256;
    const int warps_per_block = threads_per_block / 32;
    // Shared memory: space for 'outF' floats (for linear results) + one float per warp for reduction
    size_t shared_mem_bytes = (outF + warps_per_block) * sizeof(float);

    // Launch one block per sample in the batch
    fused_divergence_kernel<<<B, threads_per_block, shared_mem_bytes>>>(
        x.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.data_ptr<float>(),
        out.data_ptr<float>(),
        B,
        inF,
        outF,
        static_cast<int>(kernel_size),
        static_cast<float>(scale_factor)
    );

    return out;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Fused CUDA kernel for linear transformation and pooling");
}
